from slai_cli import log

ERROR_MAP = {
    "existing_deployments_processing": "There are existing deployments processing, please try again in a few minutes.",  # noqa
    "no_available_model_servers": "There are currently no available model servers, please try again in a few minutes.",  # noqa
    "no_artifacts_found": "No model artifacts have been saved for this model version, please save an artifact with model_version.save() and try again.",  # noqa
    "model_not_found": "Model not found, run `slai model list` to see available models in this project.",  # noqa
    "unknown_http_error": "An unknown error occurred, please reach out in our dashboard for support.",  # noqa
}


def handle_error(*, error_msg):
    log.warn(
        f"ERROR: {ERROR_MAP.get(error_msg, f'Unknown error occured: {error_msg}')}"
    )



import click
import time
import requests

from pathlib import Path
from slai.clients.model import get_model_client
from slai.clients.project import get_project_client

from slai_cli import log
from slai_cli.errors import handle_error
from slai_cli.create.local_config_helper import LocalConfigHelper
from slai_cli.modules.model_watcher import ModelWatcher
from slai_cli.modules.drive_client import DriveClient
from slai_cli.modules.docker_client import DockerClient
from slai_cli.constants import MODEL_WATCH_INTERVAL_S
from slai_cli.exceptions import RetryException


def _upload_local_notebook(local_config_helper, model_client, model_name):
    local_model_config = local_config_helper.get_local_model_config(
        model_name=model_name, model_client=model_client
    )
    model_version_id = local_model_config.get("model_version_id")

    model_notebook_google_drive_file_id = local_model_config.get(
        "model_notebook_google_drive_file_id"
    )
    model_google_drive_folder_id = local_model_config.get(
        "model_google_drive_folder_id"
    )
    cwd = Path.cwd()

    drive_client = DriveClient()
    model_notebook_google_drive_file_id = drive_client.upload_model_notebook(
        model_name=model_name,
        model_google_drive_folder_id=model_google_drive_folder_id,
        notebook_path=f"{cwd}/models/{model_name}/{model_version_id}/notebook.ipynb",
        file_id=model_notebook_google_drive_file_id,
    )

    return (
        f"https://colab.research.google.com/drive/{model_notebook_google_drive_file_id}"
    )


def open_model(model_name, watch):
    log.action("Opening notebook in browser.")

    local_config_helper = LocalConfigHelper()
    local_config_helper.check_local_config()

    project_client = get_project_client(project_name=None)
    project = project_client.get_project()
    project_name = project["name"]
    project_id = project["id"]

    try:
        model_client = get_model_client(
            model_name=model_name, project_name=project_name
        )
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 400:
            error_msg = e.response.json()["detail"]["non_field_errors"][0]
            handle_error(error_msg=error_msg)
            return None
        else:
            raise

    local_model_config = local_config_helper.get_local_model_config(
        model_name=model_name, model_client=model_client
    )

    model_version_id = local_model_config["model_version_id"]
    model_notebook_url = None
    if local_config_helper.drive_integration_enabled:
        model_notebook_url = _upload_local_notebook(
            local_config_helper=local_config_helper,
            model_client=model_client,
            model_name=model_name,
        )
    else:
        try:
            docker_client = DockerClient(
                project_name=project_name, project_id=project_id
            )
        except RuntimeError:
            log.warn("ERROR: Unable to connect to docker, is it installed/running?")
            return

        try:
            server_started = docker_client.launch_local_notebook_server(
                model_name=model_name
            )
        except RetryException:
            server_started = docker_client.launch_local_notebook_server(
                model_name=model_name
            )

        if server_started:
            model_notebook_url = f"http://localhost:{docker_client.host_port}/notebooks/models/{model_name}/{model_version_id}/notebook.ipynb"
            log.action("Waiting for notebook server to start...")
            time.sleep(5.0)
        else:
            log.warn("Failed to start local notebook server.")
            return

    if model_notebook_url:
        click.launch(f"{model_notebook_url}")
    else:
        log.warn("No remote notebook URL found.")
        return

    if watch and local_config_helper.drive_integration_enabled:
        model_watching_thread = ModelWatcher(
            model_name=model_name,
            model_client=model_client,
            google_drive_integration_enabled=local_config_helper.drive_integration_enabled,
        )
        model_watching_thread.daemon = True
        model_watching_thread.start()

        while True:
            try:
                time.sleep(MODEL_WATCH_INTERVAL_S)
            except KeyboardInterrupt:
                model_watching_thread.stop()
                model_watching_thread.join()
                break

        log.action("Goodbye.")



import click
import sys
import requests

from functools import update_wrapper

from requests.exceptions import HTTPError
from slai_cli.create import commands as create
from slai_cli.model import commands as model
from slai_cli.profile import commands as profile
from slai_cli.decorators import requires_slai_credentials
from slai_cli.errors import handle_error


def with_profile(f):
    @click.pass_context
    def run(ctx, *args, **kwargs):
        _f = requires_slai_credentials(f, **kwargs)
        return ctx.invoke(_f, *args, **kwargs)

    return update_wrapper(run, f)


@click.group()
@click.option("--profile", required=False, default="default", help="Profile to use")
@with_profile
def entry_point(profile):
    pass


def main():
    entry_point.add_command(create.create)
    entry_point.add_command(model.model)
    entry_point.add_command(profile.profile)

    try:
        entry_point()
    except HTTPError:
        handle_error(error_msg="unknown_http_error")


if __name__ == "__main__":
    sys.exit(main())



import os
import yaml

from pathlib import Path
from slai_cli import log, constants
from slai_cli.modules.drive_client import DriveClient
from slai.clients.project import get_project_client
from slai.clients.identity import get_identity_client


class LocalConfigHelper:
    def __init__(self):
        self.local_path = os.getcwd()
        self.project_client = get_project_client(project_name=None)
        self.identity_client = get_identity_client()
        self.user_info = self.identity_client.get_user()

        self.drive_integration_enabled = False
        if (
            self.user_info["identity_user_google_auth"] is not None
            and self.user_info["identity_user_google_auth"]["enabled"]
        ):
            self.drive_integration_enabled = True
            self.drive_client = DriveClient()

    def check_local_config(self):
        self.project_name = self.project_client.get_project_name()
        self.project = self.project_client.get_project()
        self.service_name = f"{self.project['name']}-{self.project['id']}"
        self.local_config = self._load_local_config()
        self.google_drive_config = self._load_google_drive_config()

    def get_local_config(self):
        local_config = None

        with open(constants.LOCAL_CONFIG_PATH, "r") as f_in:
            try:
                local_config = yaml.safe_load(f_in)
            except yaml.YAMLError:
                pass

        return local_config

    def get_google_drive_config(self):
        google_drive_config = None

        if os.path.exists(constants.GOOGLE_DRIVE_CONFIG_PATH):
            with open(constants.GOOGLE_DRIVE_CONFIG_PATH, "r") as f_in:
                try:
                    google_drive_config = yaml.safe_load(f_in)
                except yaml.YAMLError:
                    pass

        return google_drive_config

    def get_local_model_config(self, *, model_name, model_client):
        local_config = self.get_local_config()
        google_drive_config = self.get_google_drive_config()
        local_model_config = None

        if model_name not in local_config["models"].keys():
            model_version_id = model_client.model["model_version_id"]

            self.update_local_model_config(
                model_name=model_name,
                model_version_id=model_version_id,
            )

            local_config = self.get_local_config()

        if (
            self.drive_integration_enabled
            and model_name not in google_drive_config["models"].keys()
        ):
            model_version_id = local_config["models"][model_name]["model_version_id"]
            log.action("Uploading template notebook to google drive.")

            cwd = Path.cwd()

            model_google_drive_folder_id = self.drive_client.create_model_folder(
                model_name=model_name,
                project_google_drive_folder_id=google_drive_config[
                    "project_google_drive_folder_id"
                ],
            )
            model_notebook_google_drive_file_id = self.drive_client.upload_model_notebook(
                model_name=model_name,
                model_google_drive_folder_id=model_google_drive_folder_id,
                notebook_path=f"{cwd}/models/{model_name}/{model_version_id}/notebook.ipynb",
            )

            self.update_google_drive_config(
                model_name=model_name,
                model_version_id=model_version_id,
                model_google_drive_folder_id=model_google_drive_folder_id,
                model_notebook_google_drive_file_id=model_notebook_google_drive_file_id,
            )
            google_drive_config = self.get_google_drive_config()

        local_model_config = local_config["models"][model_name]

        if self.drive_integration_enabled:
            google_drive_model_config = google_drive_config["models"][model_name]
            local_model_config[
                "model_google_drive_folder_id"
            ] = google_drive_model_config["model_google_drive_folder_id"]
            local_model_config[
                "model_notebook_google_drive_file_id"
            ] = google_drive_model_config["model_notebook_google_drive_file_id"]

        return local_model_config

    def update_google_drive_config(
        self,
        *,
        model_name,
        model_version_id,
        model_google_drive_folder_id=None,
        model_notebook_google_drive_file_id=None,
    ):
        google_drive_config = self.get_google_drive_config()

        model_config = {}
        model_config["model_version_id"] = model_version_id

        if model_google_drive_folder_id is not None:
            model_config["model_google_drive_folder_id"] = model_google_drive_folder_id

        if model_notebook_google_drive_file_id is not None:
            model_config[
                "model_notebook_google_drive_file_id"
            ] = model_notebook_google_drive_file_id

        google_drive_config["models"][model_name] = model_config

        with open(constants.GOOGLE_DRIVE_CONFIG_PATH, "w") as f_out:
            yaml.dump(google_drive_config, f_out, default_flow_style=False)

    def update_local_model_config(
        self,
        *,
        model_name,
        model_version_id,
    ):
        local_config = self.get_local_config()

        model_config = {}
        model_config["model_version_id"] = model_version_id

        local_config["models"][model_name] = model_config

        with open(constants.LOCAL_CONFIG_PATH, "w") as f_out:
            yaml.dump(local_config, f_out, default_flow_style=False)

    def checkout_model_version(
        self,
        *,
        model_name,
        model_version_id,
    ):
        local_config = self.get_local_config()
        local_config["models"][model_name]["model_version_id"] = model_version_id

        with open(constants.LOCAL_CONFIG_PATH, "w") as f_out:
            yaml.dump(local_config, f_out, default_flow_style=False)

    def _load_local_config(self):
        if not os.path.exists(constants.LOCAL_CONFIG_PATH):
            log.info("No local configuration found, creating new config file.")

            config_data = {
                "models": {},
            }

            with open(constants.LOCAL_CONFIG_PATH, "w") as f_out:
                yaml.dump(config_data, f_out, default_flow_style=False)

    def _load_google_drive_config(self):
        if (
            not os.path.exists(constants.GOOGLE_DRIVE_CONFIG_PATH)
            and self.drive_integration_enabled
        ):
            log.info(
                "No google drive configuration found, creating new file and uploading configuration"
            )
            project_folder_id = self._create_google_drive_folder()

            config_data = {
                "project_google_drive_folder_id": project_folder_id,
                "models": {},
            }

            with open(constants.GOOGLE_DRIVE_CONFIG_PATH, "w") as f_out:
                yaml.dump(config_data, f_out, default_flow_style=False)

    def _create_google_drive_folder(self):
        log.action("Creating project folder in google drive")

        project_folder_id = self.drive_client.create_project_folder(
            folder_name=f"slai-{self.service_name}"
        )

        log.action("Done.")
        return project_folder_id



class InvalidModelInputException(Exception):
    pass


class InvalidModelOutputException(Exception):
    pass


class ProjectExistsException(Exception):
    pass


class ModelExistsException(Exception):
    pass


class InvalidPathException(Exception):
    pass


class InvalidApiKey(Exception):
    pass


class InvalidHandlerSyntax(Exception):
    pass


class RetryException(Exception):
    pass



import click
import sys

from pathlib import Path
from slai_cli import log
from slai_cli.create.local_config_helper import LocalConfigHelper
from slai_cli.profile.configure import get_credentials
from slai.modules.runtime import detect_credentials, detect_runtime
from slai.exceptions import NoCredentialsFound
from os import path


def requires_slai_credentials(callback, *outer_args, **outer_kwargs):
    def wrapper(*args, **kwargs):
        profile_name = outer_kwargs.get("profile", "default")
        runtime = detect_runtime()

        try:
            detect_credentials(runtime=runtime, profile_name=profile_name)
        except NoCredentialsFound:
            log.warn("No credentials detected.")

            try:
                has_api_key = click.confirm("Do you have an api key?")
                if has_api_key:
                    get_credentials(profile_name=profile_name, runtime=runtime)
                else:
                    log.action("Create one at slai.io, then re-run this command.")
                    click.launch("https://slai.io")
                    sys.exit(0)
                    return

            except click.exceptions.Abort:
                log.warn("Aborted.")
                return

        return callback(*args, **kwargs)

    return wrapper


def requires_slai_project(callback):
    def wrapper(*args, **kwargs):

        slai_directory_present = Path(".slai").is_dir()

        if not slai_directory_present:
            log.warn(
                "No project detected in this directory, create one with 'slai create <project_name>'."  # noqa
            )
            return
        elif slai_directory_present and not Path(".slai/config.yml").exists():
            log.warn(
                "No project detected in this directory, create one with 'slai create <project_name>'."  # noqa
            )
            return

        local_config_helper = LocalConfigHelper()
        local_config_helper.check_local_config()

        return callback(*args, **kwargs)

    return wrapper



import requests

from slai.clients.model import get_model_client
from slai.clients.project import get_project_client
from slai.clients.cli import get_cli_client
from slai_cli.create.local_config_helper import LocalConfigHelper
from slai_cli.modules.docker_client import DockerClient
from slai_cli.errors import handle_error
from slai_cli import log, constants


def _install_requirement_interactively(*, model_name, docker_client, requirement):
    exit_code = docker_client.install_requirement(
        model_name=model_name, requirement=requirement
    )
    if exit_code != constants.REQUIREMENTS_RETURN_CODE_SUCCESS:
        log.warn(f"Failed to install: {requirement}")


def _check_imports(*, model_name, model_version_id, docker_client):
    exit_code = docker_client.check_trainer_imports(
        model_name=model_name, model_version_id=model_version_id
    )

    if exit_code == constants.REQUIREMENTS_RETURN_CODE_MISSING_DEP:
        if log.action_confirm(
            "Looks like a dependency is missing, would you like to add it interactively?"
        ):
            requirement = log.action_prompt(
                "Add python dependency (e.g. numpy==1.20.1): ",
                type=str,
            )
            return _install_requirement_interactively(
                model_name=model_name,
                docker_client=docker_client,
                requirement=requirement,
            )
        else:
            return constants.REQUIREMENTS_RETURN_CODE_EXIT
    elif exit_code == constants.REQUIREMENTS_RETURN_CODE_FILE_NOT_FOUND_ERROR:
        log.warn(
            "Trainer script not found, run `slai model save <model_name>` to generate a training script from your notebook."  # noqa
        )
        return constants.REQUIREMENTS_RETURN_CODE_EXIT
    elif exit_code == constants.REQUIREMENTS_RETURN_CODE_IMPORT_ERROR:
        log.warn("Trainer script failed due to import error.")
        return constants.REQUIREMENTS_RETURN_CODE_EXIT

    return constants.REQUIREMENTS_RETURN_CODE_SUCCESS


def _run_trainer(*, model_name, model_version_id, docker_client):
    docker_client.create_model_environment(model_name=model_name)

    log.info("Checking if model trainer imports are valid...")
    import_return_code = _check_imports(
        model_name=model_name,
        model_version_id=model_version_id,
        docker_client=docker_client,
    )

    while (
        import_return_code != constants.REQUIREMENTS_RETURN_CODE_SUCCESS
        and import_return_code != constants.REQUIREMENTS_RETURN_CODE_EXIT  # noqa
    ):
        import_return_code = _check_imports(
            model_name=model_name,
            model_version_id=model_version_id,
            docker_client=docker_client,
        )

    if import_return_code == constants.REQUIREMENTS_RETURN_CODE_SUCCESS:
        log.info("Dependencies look valid, running training script...")
        exit_code = docker_client.run_trainer(
            model_name=model_name,
            model_version_id=model_version_id,
        )
    else:
        exit_code = import_return_code

    if exit_code != constants.REQUIREMENTS_RETURN_CODE_SUCCESS:
        log.warn(f"Trainer script failed with return code: {exit_code}")
        log.warn("Training failed.")
        return False
    else:
        log.action("Training complete.")
        return True


def train_model(name):
    log.action(f"Training model: {name}")

    project_client = get_project_client(project_name=None)
    project = project_client.get_project()
    project_name = project["name"]
    project_id = project["id"]

    cli_client = get_cli_client()
    try:
        docker_client = DockerClient(project_name=project_name, project_id=project_id)
    except RuntimeError:
        log.warn("ERROR: Unable to connect to docker, is it installed/running?")
        return

    try:
        model_client = get_model_client(model_name=name, project_name=project_name)
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 400:
            error_msg = e.response.json()["detail"]["non_field_errors"][0]
            handle_error(error_msg=error_msg)
            return None
        else:
            raise

    local_config_helper = LocalConfigHelper()
    local_config = local_config_helper.get_local_config()

    try:
        model_version_id = local_config["models"][name]["model_version_id"]
    except KeyError:
        log.action("No local config set, using default model version.")
        model_version_id = model_client.model["model_version_id"]

    model_version = cli_client.retrieve_model_version_by_id(
        model_version_id=model_version_id
    )

    log.action(f"Using model version: {model_version['name']} <{model_version_id}>")
    if not log.warn_confirm(
        "If your training script saves a model artifact, this will be uploaded to the slai backend, continue?",  # noqa
    ):
        return

    _ = _run_trainer(
        model_name=name,
        model_version_id=model_version_id,
        docker_client=docker_client,
    )



import yaml

from columnar import columnar
from pathlib import Path


from slai_cli import log


def list_profiles():
    headers = ["name", "client_id", "client_secret"]

    credentials_path = f"{Path.home()}/.slai"
    try:
        with open(f"{credentials_path}/credentials.yml", "r") as f_in:
            try:
                credentials = yaml.safe_load(f_in)
            except yaml.YAMLError:
                pass
    except:
        credentials = {}

    data = [
        [
            profile_name,
            credentials[profile_name]["client_id"],
            credentials[profile_name]["client_secret"],
        ]
        for profile_name in credentials.keys()
    ]
    try:
        table = columnar(data, headers, no_borders=True)
        log.info(table)
    except IndexError:
        log.warn("No profiles found.")
        return



from slai_cli import log
from slai_cli import constants


def _install_requirement_interactively(*, model_name, docker_client, requirement):
    exit_code = docker_client.install_requirement(
        model_name=model_name, requirement=requirement
    )
    if exit_code != constants.REQUIREMENTS_RETURN_CODE_SUCCESS:
        log.warn(f"Failed to install: {requirement}")


def _check_handler_imports(*, model_name, docker_client):
    exit_code = docker_client.check_handler_imports(
        model_name=model_name,
    )
    if exit_code == constants.REQUIREMENTS_RETURN_CODE_MISSING_DEP:
        if log.action_confirm(
            "Looks like a dependency is missing, would you like to add it interactively?"
        ):
            requirement = log.action_prompt(
                "Add python dependency (e.g. numpy==1.20.1): ",
                type=str,
            )

            return _install_requirement_interactively(
                model_name=model_name,
                docker_client=docker_client,
                requirement=requirement,
            )
        else:
            return constants.REQUIREMENTS_RETURN_CODE_EXIT

    elif exit_code == constants.REQUIREMENTS_RETURN_CODE_IMPORT_ERROR:
        log.warn("Trainer script failed due to import error.")
        return constants.REQUIREMENTS_RETURN_CODE_IMPORT_ERROR

    return constants.REQUIREMENTS_RETURN_CODE_SUCCESS


def check_handler_imports(*, model_name, docker_client):
    log.action("Checking if model handler imports are valid...")

    import_return_code = _check_handler_imports(
        model_name=model_name, docker_client=docker_client
    )
    while (
        import_return_code != constants.REQUIREMENTS_RETURN_CODE_SUCCESS
        and import_return_code != constants.REQUIREMENTS_RETURN_CODE_EXIT
    ):
        import_return_code = _check_handler_imports(
            model_name=model_name,
            docker_client=docker_client,
        )

    return import_return_code



import time

from pathlib import Path
from slai.modules.notebook_utils import generate_trainer_from_template, parse_notebook
from slai_cli import log
from slai_cli.create.local_config_helper import LocalConfigHelper


def save_model(model_name):
    log.action(f"Saving current model notebook state on {time.ctime()}")

    local_config_helper = LocalConfigHelper()
    local_config_helper.check_local_config()

    local_config = local_config_helper.get_local_config()
    model_version_id = local_config["models"][model_name]["model_version_id"]

    cwd = Path.cwd()
    notebook_path = f"{cwd}/models/{model_name}/{model_version_id}/notebook.ipynb"

    imports, trainer = parse_notebook(notebook_path)
    trainer_source = generate_trainer_from_template(
        model_name=model_name,
        imports=imports,
        trainer=trainer,
        model_version_id=model_version_id,
    )

    cwd = Path.cwd()
    with open(f"{cwd}/models/{model_name}/{model_version_id}/trainer.py", "w") as f_out:
        f_out.write(trainer_source)

    log.action("Done.\n")



import click


OUTPUT_PREFIX = "slai: "


def warn(str):
    click.echo(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='red')}"
    )


def info(str):
    click.echo(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='white')}"
    )


def action(str):
    click.echo(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='yellow')}"
    )


def warn_confirm(str):
    return click.confirm(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='red')}",
    )


def action_confirm(_str):
    return click.confirm(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(_str, bg='black', fg='yellow')}",
    )


def action_prompt(str, type, default=""):
    return click.prompt(
        f"{click.style(OUTPUT_PREFIX, bg='black', fg='white')}{click.style(str, bg='black', fg='yellow')}",
        type=type,
        show_default=True,
        default=default,
    )



import click


from slai_cli.profile import configure as _configure
from slai_cli.profile import list as _list
from slai.modules.runtime import ValidRuntimes


@click.group()
def profile():
    pass


@click.argument("profile_name")
@click.option("--client-id", required=False, help="Client ID.")
@click.option("--client-secret", required=False, help="Client Secret.")
@profile.command()
def configure(profile_name, client_id, client_secret):
    _configure.get_credentials(
        profile_name=profile_name,
        runtime=ValidRuntimes.Local,
        client_id=client_id,
        client_secret=client_secret,
    )


@profile.command()
def list():
    _list.list_profiles()



from columnar import columnar

from slai.clients.project import get_project_client

from slai_cli import log
from slai_cli.decorators import requires_slai_project


@requires_slai_project
def list_models():
    project_client = get_project_client(project_name=None)
    models = project_client.list_models()

    headers = ["id", "name", "created", "updated", "version"]
    data = [
        [
            model["id"],
            model["name"],
            model["created"],
            model["updated"],
            model["model_version_id"],
        ]
        for model in models
    ]
    try:
        table = columnar(data, headers, no_borders=True)
        log.info(table)
    except IndexError:
        log.warn("No models found.")
        return



from slai.clients.gdrive import get_google_drive_client


class DriveClient:
    def __init__(self):
        self.client = get_google_drive_client()

    def upload_model_notebook(
        self, model_name, model_google_drive_folder_id, notebook_path, file_id=None
    ):
        model_notebook_google_drive_file_id = self.client.upload_file(
            filename=f"{model_name}.ipynb",
            local_path=notebook_path,
            parent_ids=[
                model_google_drive_folder_id,
            ],
            file_id=file_id,
        )

        return model_notebook_google_drive_file_id

    def download_latest_model_notebook(
        self, local_path, model_notebook_google_drive_file_id
    ):
        self.client.download_file(
            file_id=model_notebook_google_drive_file_id,
            local_path=local_path,
        )

    def create_model_folder(self, model_name, project_google_drive_folder_id):
        model_google_drive_folder_id = self.client.create_folder(
            name=f"{model_name}",
            parent_ids=[project_google_drive_folder_id],
        )
        return model_google_drive_folder_id

    def create_project_folder(self, folder_name):
        project_folder_google_drive_id = self.client.create_folder(
            name=f"{folder_name}",
            parent_ids=[],
        )
        return project_folder_google_drive_id



import os
import shutil
import requests
import git

from columnar import columnar
from pathlib import Path
from git import Repo

from slai.clients.model import get_model_client
from slai.clients.cli import get_cli_client
from slai.clients.project import get_project_client
from slai_cli.errors import handle_error
from slai_cli.create.local_config_helper import LocalConfigHelper
from slai_cli import log


def _list_model_versions(model_versions):
    headers = ["index", "name", "id", "created"]
    data = [
        [
            index,
            mv["name"],
            mv["id"],
            mv["created"],
        ]
        for index, mv in enumerate(model_versions)
    ]
    try:
        table = columnar(data, headers, no_borders=True)
        log.info(table)
    except IndexError:
        log.warn("No model versions found.")
        return


def _create_local_version_path(
    *, model_name, new_model_version_id, previous_model_version_id, new_version
):
    log.action("Copying existing model notebook...")

    cwd = Path.cwd()
    os.makedirs(f"{cwd}/models/{model_name}/{new_model_version_id}")

    notebook_to_copy = Path(
        f"{cwd}/models/{model_name}/{previous_model_version_id}/notebook.ipynb"
    )
    new_notebook = Path(
        f"{cwd}/models/{model_name}/{new_model_version_id}/notebook.ipynb"
    )

    shutil.copy(notebook_to_copy, new_notebook)
    with open(new_notebook) as f_in:
        _text = f_in.read().replace(previous_model_version_id, new_model_version_id)

    with open(new_notebook, "w") as f_out:
        f_out.write(_text)


def _commit_model_version(*, model_name, model_version_id, model_version_name):
    cwd = Path.cwd()
    repo = Repo.init(cwd)
    repo.git.add([f"models/{model_name}/{model_version_id}"])
    repo.git.add([f"handlers/{model_name}/*"])
    repo.index.commit(f"created model version: {model_name}/{model_version_name}")

    try:
        log.action("Pushing state to git...")
        repo.git.push("origin", "master")
        log.action("Pushed to git.")
    except (ValueError, git.exc.GitCommandError):
        log.warn(
            "WARNING: No remote is configured for this project, not pushing commits."
        )


def checkout_model_version(model_name, version_name=None):
    log.action(f"Retrieving model versions for model: {model_name}")

    project_client = get_project_client(project_name=None)
    project_name = project_client.get_project_name()

    try:
        model_client = get_model_client(
            model_name=model_name, project_name=project_name
        )
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 400:
            error_msg = e.response.json()["detail"]["non_field_errors"][0]
            handle_error(error_msg=error_msg)
            return None
        else:
            raise

    cli_client = get_cli_client()

    model_id = model_client.model["id"]
    model_versions = cli_client.list_model_versions(model_id=model_id)
    version_names = {mv["name"]: mv["id"] for mv in model_versions}

    local_config_helper = LocalConfigHelper()
    local_model_config = local_config_helper.get_local_model_config(
        model_name=model_name, model_client=model_client
    )
    model_version_id = local_model_config.get("model_version_id")
    if model_version_id is None:
        model_version_id = model_client.model["model_version_id"]

    model_version = cli_client.retrieve_model_version_by_id(
        model_version_id=model_version_id
    )

    log.action(
        f"Current model version is: {model_version['name']} <{model_version_id}>"
    )

    # if a version name was specified, check if a version with that name exists
    if version_name is not None:

        # if so, just update local_config to point to that version
        if version_name in version_names:
            model_version_id_to_checkout = version_names[version_name]
            local_config_helper.checkout_model_version(
                model_name=model_name, model_version_id=model_version_id_to_checkout
            )

            log.action(f"Checked out model version: {version_name}")

        # otherwise, create a new model version with that name
        else:
            model_version = cli_client.create_model_version(
                model_id=model_id, name=version_name
            )
            model_version_id_to_checkout = model_version["id"]
            local_config_helper.checkout_model_version(
                model_name=model_name, model_version_id=model_version_id_to_checkout
            )

            _create_local_version_path(
                model_name=model_name,
                new_model_version_id=model_version_id_to_checkout,
                previous_model_version_id=model_version_id,
                new_version=model_version,
            )

            _commit_model_version(
                model_name=model_name,
                model_version_id=model_version["id"],
                model_version_name=model_version["name"],
            )

            log.action(f"Checked out model version: {model_version['name']}")

    # if no version was specified, allow a user to checkout from existing model versions
    else:
        _list_model_versions(model_versions)
        model_version_index = log.action_prompt(
            "Checkout a version by index", type=int, default=0
        )
        if model_version_index < len(model_versions):
            model_version = model_versions[model_version_index]
            log.action(f"Checked out model version: {model_version['name']}")

            model_version_id_to_checkout = model_version["id"]
            local_config_helper.checkout_model_version(
                model_name=model_name, model_version_id=model_version_id_to_checkout
            )
        else:
            log.warn("Invalid model version.")



import click
import yaml
import os
import slai
import sys

from pathlib import Path
from slai_cli import log
from slai_cli.exceptions import InvalidApiKey
from slai.modules.runtime import ValidRuntimes


def get_credentials(
    profile_name="default",
    runtime=ValidRuntimes.Project,
    client_id=None,
    client_secret=None,
):
    if runtime == ValidRuntimes.Local:
        log.action(f"Configuring profile: {profile_name}")

    if client_id is None or client_secret is None:
        client_id = click.prompt("Client ID", type=str, default=client_id)
        client_secret = click.prompt("Client Secret", type=str, default=client_secret)

    try:
        store_credentials(
            profile_name=profile_name,
            client_id=client_id,
            client_secret=client_secret,
            runtime=runtime,
        )
    except InvalidApiKey:
        log.warn("Invalid credentials.")
        return

    log.action("Credentials configured.")


def store_credentials(*, profile_name, client_id, client_secret, runtime):
    new_profile = {
        "client_id": client_id,
        "client_secret": client_secret,
    }

    try:
        slai.login(client_id=client_id, client_secret=client_secret, key_type="USER")
    except slai.exceptions.InvalidCredentials:
        log.warn("Invalid credentials")
        sys.exit(1)

    if runtime == ValidRuntimes.Local:
        credentials_path = f"{Path.home()}/.slai"
        if not os.path.exists(credentials_path):
            os.makedirs(credentials_path)
    elif runtime == ValidRuntimes.Project:
        credentials_path = f"{Path.cwd()}/.slai"

    try:
        with open(f"{credentials_path}/credentials.yml", "r") as f_in:
            try:
                credentials = yaml.safe_load(f_in)
            except yaml.YAMLError:
                pass
    except:
        credentials = {}

    # save new profile
    if runtime == ValidRuntimes.Local:
        if credentials.get("default") is None:
            credentials["default"] = new_profile

        credentials[profile_name] = new_profile
    elif runtime == ValidRuntimes.Project:
        credentials = new_profile

    noalias_dumper = yaml.dumper.SafeDumper
    noalias_dumper.ignore_aliases = lambda self, data: True

    with open(f"{credentials_path}/credentials.yml", "w") as f_out:
        yaml.dump(credentials, f_out, default_flow_style=False, Dumper=noalias_dumper)



import time
import slai

from slai.clients.project import get_project_client
from slai_cli import log


def call_model(*, model_name, input):
    project_client = get_project_client()

    project_name = project_client.get_project()["name"]

    _model = slai.model(f"{project_name}/{model_name}")

    for i in range(100):
        start = time.time()
        response = _model(weight=1.0, age=0.1, height=67.0)
        elapsed = time.time() - start

        log.info(str(response))
        log.info(f"inference time: {elapsed}")



import time
import click

from pathlib import Path

from slai_cli import log
from slai_cli.modules.drive_client import DriveClient
from slai_cli.modules.stoppable_thread import StoppableThread
from slai_cli.constants import MODEL_WATCH_INTERVAL_S
from slai_cli.model.save import save_model
from slai_cli.create.local_config_helper import LocalConfigHelper

CYCLES_PER_DOWNLOAD = 5


class ModelWatcher(StoppableThread):
    def __init__(self, *args, **kwargs):
        self.model_name = kwargs.pop("model_name")
        self.model_client = kwargs.pop("model_client")
        self.google_drive_integration_enabled = kwargs.pop(
            "google_drive_integration_enabled"
        )
        self._cleanup()

        super(ModelWatcher, self).__init__(*args, **kwargs)

    def _cleanup(self):
        self.idx = 0

    def pull_current_notebook(self, *, model_name, model_client):
        log.action("Downloading notebook from google drive.")

        local_config_helper = LocalConfigHelper()
        local_model_config = local_config_helper.get_local_model_config(
            model_name=model_name, model_client=model_client
        )
        model_version_id = local_model_config["model_version_id"]

        # if google drive integration is enabled, download the working notebook from drive
        if local_config_helper.drive_integration_enabled:
            cwd = Path.cwd()

            drive_client = DriveClient()
            drive_client.download_latest_model_notebook(
                local_path=f"{cwd}/models/{model_name}/{model_version_id}/notebook.ipynb",
                model_notebook_google_drive_file_id=local_model_config[
                    "model_notebook_google_drive_file_id"
                ],
            )

            log.action("Done.")

        return model_version_id

    def run(self):
        while not self.stopped():
            if self.idx % CYCLES_PER_DOWNLOAD == 0:
                click.clear()

                if self.google_drive_integration_enabled:
                    self.pull_current_notebook(
                        model_name=self.model_name, model_client=self.model_client
                    )

                save_model(model_name=self.model_name)
                self.idx = 0

            self.idx += 1
            time.sleep(MODEL_WATCH_INTERVAL_S)

        self._cleanup()



TEMPLATE_REPO_URLS = [
    "git@github.com:slai-labs/slai_template.git",
    "https://github.com/slai-labs/slai_template.git",
]
MODEL_WATCH_INTERVAL_S = 1
LOCAL_CONFIG_PATH = ".slai/local_config.yml"
GOOGLE_DRIVE_CONFIG_PATH = ".slai/google_drive_config.yml"
REQUIREMENTS_RETURN_CODE_SUCCESS = 0
REQUIREMENTS_RETURN_CODE_MISSING_DEP = 1
REQUIREMENTS_RETURN_CODE_IMPORT_ERROR = 2
REQUIREMENTS_RETURN_CODE_FILE_NOT_FOUND_ERROR = 3
REQUIREMENTS_RETURN_CODE_UNKNOWN_ERROR = 4
REQUIREMENTS_RETURN_CODE_EXIT = 5



import os
import datetime
import git

from pathlib import Path
from jinja2 import Template
from requests.exceptions import HTTPError
from git import Repo

from slai.clients.project import get_project_client
from slai.clients.cli import get_cli_client

from slai_cli import log
from slai_cli.decorators import requires_slai_project
from slai_cli.exceptions import ModelExistsException, InvalidApiKey


def _create_model_folder_structure(model):
    log.action(f"Generating folder structure for model: {model['name']} ")

    cwd = Path.cwd()
    if not os.path.exists(f"{cwd}/models/{model['name']}/"):
        os.makedirs(f"{cwd}/models/{model['name']}")
        os.makedirs(f"{cwd}/models/{model['name']}/{model['model_version_id']}")
        os.makedirs(f"{cwd}/handlers/{model['name']}")
    else:
        raise ModelExistsException("model_already_exists")

    log.action("Done.")


def _write_template_file(*, populated_templates, path, filename):
    with open(f"{path}/{filename}", "w") as f_out:
        f_out.write(populated_templates[filename])


def _populate_template_files(model):
    populated_templates = {}
    template_files = ["notebook.ipynb", "handler.py", "test_event.json"]

    pwd = Path(__file__).parent

    template_variables = {
        "SLAI_MODEL_NAME": model["name"],
        "SLAI_MODEL_CREATED_AT": datetime.datetime.now().isoformat(),
        "SLAI_MODEL_VERSION_ID": model["model_version_id"],
        "SLAI_MODEL_ID": model["id"],
        "SLAI_PROJECT_NAME": model["project_name"],
        "SLAI_PROJECT_ID": model["project_id"],
    }

    # populate model template files
    for fname in template_files:
        log.action(f"Generating: {fname} ")
        template_contents = None

        with open(f"{pwd}/templates/{fname}", "r") as f_in:
            template_contents = f_in.read()
            t = Template(template_contents)
            rendered_template = t.render(**template_variables)

            populated_templates[fname] = rendered_template
            log.action("Done.")

    cwd = Path.cwd()
    _write_template_file(
        populated_templates=populated_templates,
        path=f"{cwd}/models/{model['name']}/{model['model_version_id']}",
        filename="notebook.ipynb",
    )

    _write_template_file(
        populated_templates=populated_templates,
        path=f"{cwd}/handlers/{model['name']}",
        filename="handler.py",
    )

    _write_template_file(
        populated_templates=populated_templates,
        path=f"{cwd}/handlers/{model['name']}",
        filename="test_event.json",
    )


def _commit_model(*, model_name):
    cwd = Path.cwd()
    repo = Repo.init(cwd)
    repo.git.add([f"models/{model_name}/*"])
    repo.git.add([f"handlers/{model_name}/*"])
    repo.index.commit(f"created new model: {model_name}")

    try:
        log.action("Pushing state to git...")
        repo.git.push("origin", "master")
        log.action("Pushed to git.")
    except (ValueError, git.exc.GitCommandError):
        log.warn(
            "WARNING: No remote is configured for this project, not pushing commits."
        )


def _validate_model_name(name):
    raise NotImplementedError


@requires_slai_project
def create_new_model(name, from_model=None):
    project_client = get_project_client(project_name=None)
    project_name = project_client.get_project_name()

    cli_client = get_cli_client()
    try:
        _validate_model_name(name)
    except Exception:
        pass

    try:
        model = cli_client.create_model(project_name=project_name, name=name)
    except HTTPError as e:
        if e.response.status_code == 400:
            error_msg = e.response.json()["detail"]["non_field_errors"][0]
            if error_msg == "duplicate_model_name":
                log.warn(f"Model already exists with name '{name}'")
                return False
            else:
                raise
        elif e.response.status_code == 401:
            log.warn("Invalid client id/secret")
            raise InvalidApiKey

    try:
        _create_model_folder_structure(model=model)
    except ModelExistsException:
        log.warn(f"Invalid model name, model '{name}' already exists.")
        raise

    # fill out all the templates
    _populate_template_files(model=model)

    # commit new model to git repo
    _commit_model(model_name=name)

    return True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['slai_cli',
 'slai_cli.create',
 'slai_cli.model',
 'slai_cli.model.templates',
 'slai_cli.modules',
 'slai_cli.profile']

package_data = \
{'': ['*'], 'slai_cli.create': ['templates/*']}

install_requires = \
['Columnar>=1.3.1,<2.0.0',
 'Flask-Cors>=3.0.9,<4.0.0',
 'Flask>=1.1.2,<2.0.0',
 'GitPython>=3.1.11,<4.0.0',
 'Jinja2>=2.11.2,<3.0.0',
 'PyDrive>=1.3.1,<2.0.0',
 'boto3>=1.16.44,<2.0.0',
 'celery==5.0.2',
 'click>=7.1.2,<8.0.0',
 'dill>=0.3.3,<0.4.0',
 'docker>=5.0.0,<6.0.0',
 'python-dotenv>=0.15.0,<0.16.0',
 'redis>=3.5.3,<4.0.0',
 'shtab>=1.3.4,<2.0.0',
 'slai==0.1.29']

entry_points = \
{'console_scripts': ['slai = slai_cli:main.main']}

setup_kwargs = {
    'name': 'slai-cli',
    'version': '0.4.8',
    'description': '',
    'long_description': None,
    'author': 'slai',
    'author_email': 'luke@slai.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0.0',
}


setup(**setup_kwargs)



import click

from slai.modules.runtime import detect_runtime, detect_credentials
from slai.exceptions import NoCredentialsFound
from slai_cli import log
from slai_cli.create import project_generator
from slai_cli.exceptions import ProjectExistsException, InvalidApiKey


@click.command()
@click.argument("name")
@click.option("--profile", required=False, default="default", help="Profile to use")
def create(name, profile):
    """Create a new slai project."""

    runtime = detect_runtime()

    try:
        credentials = detect_credentials(runtime=runtime, profile_name=profile)
    except NoCredentialsFound:
        log.warn("Invalid credentials.")
        return

    try:
        pg = project_generator.ProjectGenerator(
            project_name=name,
            credentials=credentials,
        )
    except ProjectExistsException:
        log.warn(f"Project '{name}' already exists.")
        return
    except InvalidApiKey:
        log.warn("Invalid credentials.")
        return

    pg.create_new_project()



import base64
import json
import requests
import git

from pathlib import Path
from requests.exceptions import HTTPError
from git import Repo

from slai_cli import log
from slai_cli.modules.deployment_checks import check_handler_imports
from slai_cli.decorators import requires_slai_project
from slai_cli.exceptions import InvalidHandlerSyntax
from slai_cli.modules.docker_client import DockerClient
from slai_cli import constants
from slai_cli.errors import handle_error
from slai_cli.create.local_config_helper import LocalConfigHelper

from slai.clients.cli import get_cli_client
from slai.clients.model import get_model_client
from slai.clients.project import get_project_client


def _merge_requirements(model_artifact_requirements, model_handler_requirements):
    requirements = {}
    for req, version in model_artifact_requirements.items():
        requirements[req] = version

    for req, version in model_handler_requirements.items():
        requirements[req] = version

    return requirements


def _create_model_deployment(*, model_client, model_version_id, docker_client):
    cwd = Path.cwd()
    model = model_client.model
    model_name = model["name"]

    handler_path = f"{cwd}/handlers/{model['name']}/handler.py"
    log.action(f"Deploying model handler stored at: {handler_path}")
    log.warn(
        "Currently, slai can only check for syntax/import errors, ensure your handler script has been tested!"  # noqa
    )

    # check for syntax errors in the model handler
    try:
        handler_source = open(handler_path, "r").read() + "\n"
        compile(handler_source, handler_path, "exec")
    except SyntaxError:
        raise InvalidHandlerSyntax("invalid_model_handler")

    handler_data = base64.b64encode(handler_source.encode("utf-8")).decode()

    try:
        model_artifact = model_client.get_latest_model_artifact(
            model_version_id=model_version_id
        )
    except HTTPError as e:
        if e.response.status_code == 400:
            error_msg = e.response.json()["detail"]["non_field_errors"][0]
            handle_error(error_msg=error_msg)
            return None
        else:
            raise

    # check model requirements in local container
    model_artifact_requirements = {}
    docker_client.create_model_environment(model_name=model_name)
    import_return_code = docker_client.check_model_requirements(
        model_name=model_name, requirements=model_artifact["artifact_requirements"]
    )
    if import_return_code == constants.REQUIREMENTS_RETURN_CODE_SUCCESS:
        model_artifact_requirements = model_artifact["artifact_requirements"]

    # check if model handler imports are valid
    import_return_code = check_handler_imports(
        model_name=model_name, docker_client=docker_client
    )

    if import_return_code != constants.REQUIREMENTS_RETURN_CODE_SUCCESS:
        log.warn("Deployment failed due to import error.")
        return None

    log.action("Model handler imports look valid.")
    model_handler_requirements = docker_client.generate_model_handler_requirements(
        model_name=model_name
    )

    requirements = _merge_requirements(
        model_artifact_requirements, model_handler_requirements
    )

    # requirements look valid, deploy model
    log.action("Deploying model...")
    try:
        model_deployment = model_client.create_model_deployment(
            model_artifact_id=model_artifact["id"],
            model_handler_data=handler_data,
            requirements=requirements,
        )
    except HTTPError as e:
        if e.response.status_code == 400:
            error_msg = e.response.json()["detail"][0]
            handle_error(error_msg=error_msg)
            return None
        else:
            raise

    return model_deployment


def _commit_deployment(*, model_name, model_version_name):
    cwd = Path.cwd()
    repo = Repo.init(cwd)
    repo.git.add([f"models/{model_name}/*"])
    repo.git.add([f"handlers/{model_name}/*"])
    repo.index.commit(f"deployed model: {model_name}/{model_version_name}")

    try:
        log.action("Pushing state to git...")
        repo.git.push("origin", "master")
        log.action("Pushed to git.")
    except (ValueError, git.exc.GitCommandError):
        log.warn(
            "WARNING: No remote is configured for this project, not pushing commits."
        )


@requires_slai_project
def deploy_model(*, model_name, version):
    log.action("Deploying model to slai backend.")

    project_client = get_project_client(project_name=None)
    project = project_client.get_project()
    project_name = project["name"]
    project_id = project["id"]

    cli_client = get_cli_client()

    try:
        docker_client = DockerClient(project_name=project_name, project_id=project_id)
    except RuntimeError:
        log.warn("ERROR: Unable to connect to docker, is it installed/running?")
        return

    try:
        model_client = get_model_client(
            model_name=model_name, project_name=project_name
        )
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 400:
            error_msg = e.response.json()["detail"]["non_field_errors"][0]
            handle_error(error_msg=error_msg)
            return None
        else:
            raise

    local_config_helper = LocalConfigHelper()
    local_model_config = local_config_helper.get_local_model_config(
        model_name=model_name, model_client=model_client
    )

    model_version_id = local_model_config.get("model_version_id")
    if model_version_id is None:
        model_version_id = model_client.model["model_version_id"]

    model_version = cli_client.retrieve_model_version_by_id(
        model_version_id=model_version_id
    )

    log.action(f"Using model version: {model_version['name']} <{model_version_id}>")

    try:
        model_deployment = _create_model_deployment(
            model_client=model_client,
            model_version_id=model_version_id,
            docker_client=docker_client,
        )
        if model_deployment is not None:
            log.action(
                f"Created new model deployment:\n{json.dumps(model_deployment, sort_keys=True, indent=4)}"  # noqa
            )
            log.action("Done.")

            # commit changes to git repo
            _commit_deployment(
                model_name=model_name, model_version_name=model_version["name"]
            )

    except InvalidHandlerSyntax:
        log.warn(
            "ERROR: Cannot deploy a model handler with syntax/import errors, please check your handler.py"  # noqa
        )



import threading


class StoppableThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()



import requests

from slai.clients.project import get_project_client
from slai.clients.cli import get_cli_client
from slai.clients.model import get_model_client

from slai_cli import log
from slai_cli.errors import handle_error
from slai_cli.modules.docker_client import DockerClient
from slai_cli.modules.deployment_checks import check_handler_imports
from slai_cli.create.local_config_helper import LocalConfigHelper
from slai_cli import constants


def _run_inference(*, model_name, model_version_id, docker_client):
    docker_client.create_model_environment(model_name=model_name)

    exit_code = docker_client.test_model(
        model_name=model_name,
        model_version_id=model_version_id,
    )

    return exit_code


def test_model(*, model_name):
    log.action(f"Testing model: {model_name}")

    project_client = get_project_client(project_name=None)
    project = project_client.get_project()
    project_name = project["name"]
    project_id = project["id"]
    cli_client = get_cli_client()

    try:
        docker_client = DockerClient(project_name=project_name, project_id=project_id)
    except RuntimeError:
        log.warn("ERROR: Unable to connect to docker, is it installed/running?")
        return

    try:
        model_client = get_model_client(
            model_name=model_name, project_name=project_name
        )
    except requests.exceptions.HTTPError as e:
        if e.response.status_code == 400:
            error_msg = e.response.json()["detail"]["non_field_errors"][0]
            handle_error(error_msg=error_msg)
            return None
        else:
            raise

    local_config_helper = LocalConfigHelper()
    local_config = local_config_helper.get_local_config()

    try:
        model_version_id = local_config["models"][model_name]["model_version_id"]
    except KeyError:
        log.action("No local config set, using default model version.")
        model_version_id = model_client.model["model_version_id"]

    model_version = cli_client.retrieve_model_version_by_id(
        model_version_id=model_version_id
    )

    log.action(f"Using model version: {model_version['name']} <{model_version_id}>")

    # check if model handler imports are valid
    import_return_code = check_handler_imports(
        model_name=model_name, docker_client=docker_client
    )
    if import_return_code != constants.REQUIREMENTS_RETURN_CODE_SUCCESS:
        log.warn("Failed failed due to import error.")
        return None

    _ = _run_inference(
        model_name=model_name,
        model_version_id=model_version_id,
        docker_client=docker_client,
    )



import click

from slai_cli.model.create import create_new_model
from slai_cli.model.save import save_model
from slai_cli.model.open import open_model
from slai_cli.model.list import list_models
from slai_cli.model.train import train_model
from slai_cli.model.call import call_model
from slai_cli.model.test import test_model
from slai_cli.model.deploy import deploy_model
from slai_cli.model.checkout import checkout_model_version
from slai_cli.exceptions import ModelExistsException, InvalidApiKey


@click.group()
def model():
    pass


@model.command()
@click.argument("model_name")
@click.option("--template", required=False, help="An optional notebook template.")
@click.option(
    "--from-model", required=False, help="Start from another model in your project"
)
@click.option("--watch/--no-watch", default=True)
def create(model_name, template, from_model, watch):
    """Create new model."""

    try:
        created = create_new_model(model_name, from_model=from_model)
    except (ModelExistsException, InvalidApiKey):
        return

    if created:
        open_model(model_name, watch=watch)


@model.command()
@click.argument("model_name")
def save(model_name):
    """Save changes to local model trainer."""
    save_model(model_name)


@model.command()
@click.argument("model_name")
@click.option("--watch/--no-watch", default=True)
def open(model_name, watch):
    """Open model notebook in your browser."""
    open_model(model_name, watch=watch)


@model.command()
@click.argument("model_name")
def train(model_name):
    """Train a model locally using the local trainer files."""
    train_model(model_name)


@model.command()
@click.argument("model_name")
@click.option("-name", default=None, required=False)
def checkout(model_name, name):
    checkout_model_version(model_name=model_name, version_name=name)


@model.command()
@click.argument("model_name")
def call(model_name):
    """Call a model using the inference client directly."""
    call_model(model_name=model_name, input={})


@model.command()
@click.argument("model_name")
def test(model_name):
    """Test a model/handler locally before deploying."""
    test_model(model_name=model_name)


@model.command()
@click.argument("model_name")
@click.option(
    "--version", required=False, default="latest", help="An option model version ID."
)
def deploy(model_name, version):
    """
    Deploy a model artifact and handler.
    """

    # TODO: support explicit model artifact IDs
    deploy_model(
        model_name=model_name,
        version=version,
    )


@model.command()
def list():
    """List models in the project."""
    list_models()



import os
import click
import git
import yaml
import shutil

from jinja2 import Template
from pathlib import Path
from shutil import rmtree
from git import Repo
from requests.exceptions import HTTPError

from slai_cli import log
from slai.clients.cli import get_cli_client

from slai_cli.constants import TEMPLATE_REPO_URLS
from slai_cli.exceptions import (
    InvalidPathException,
    ProjectExistsException,
    InvalidApiKey,
)


class ProjectGenerator:
    def __init__(
        self,
        *,
        project_name,
        credentials,
    ):
        self.cli_client = get_cli_client(
            client_id=credentials["client_id"],
            client_secret=credentials["client_secret"],
        )
        try:
            self.cli_client.get_user()
        except HTTPError as e:
            if e.response.status_code == 401:
                raise InvalidApiKey("invalid_credentials")
            else:
                raise

        self.credentials = credentials
        self.project_name = project_name

    def create_new_project(self):
        click.clear()
        log.action(f"Creating a new project with name: {self.project_name}")

        slai_directory_present = Path(".slai").is_dir()
        if slai_directory_present and Path(".slai/config.yml").exists():
            log.warn(
                "You cannot create nested slai projects, try again in another directory."  # noqa
            )
            return

        try:
            local_path = self._create_project_directory()
            os.chdir(local_path)
        except InvalidPathException as e:
            log.warn(f"Invalid project name: {e}\n")
            return

        try:
            # clone template repository
            self._clone_template_repo(local_path=local_path)
        except (RuntimeError):
            return

        try:
            self.project_variables = self._generate_project_variables(self.project_name)
        except (ProjectExistsException, InvalidPathException):
            log.warn(
                f"A project or directory already exists with name '{self.project_name}', aborting."
            )
            project_path = Path(local_path)
            shutil.rmtree(project_path)
            return

        # store profile in project credentials file
        self._store_local_credentials()

        # generate template files
        self._create_project_files(local_path=local_path)

        # initialize git repo and commit
        self._init_git_repo(local_path)

    def _init_git_repo(self, local_path):
        repo = Repo.init(local_path)
        repo.git.add("--all")
        repo.index.commit("initial commit")

    def _create_project_directory(self):
        cwd = os.getcwd()
        local_path = f"{cwd}/{self.project_name}"
        if not os.path.exists(local_path):
            os.makedirs(local_path)

        if os.path.exists(f"{cwd}/{self.project_name}/.slai"):
            raise InvalidPathException("project_already_exists")

        return local_path

    def _store_local_credentials(self):
        cwd = os.getcwd()

        with open(f"{cwd}/.slai/credentials.yml", "w") as f_out:
            yaml.dump(self.credentials, f_out, default_flow_style=False)

    def _generate_project_variables(self, project_name):
        cwd = os.getcwd()
        if os.path.exists(f"{cwd}/{self.project_name}"):
            raise InvalidPathException("directory_exists")

        try:
            self.project = self.cli_client.create_project(name=self.project_name)
        except HTTPError as e:
            if e.response.status_code == 400:
                raise ProjectExistsException("duplicate_project_name")
            elif e.response.status_code == 401:
                raise InvalidApiKey("invalid_credentials")
            else:
                raise

        self.service_name = f"{project_name}-{self.project['id']}"

        variables = {
            "SLAI_SERVICE_NAME": self.service_name,
            "SLAI_PROJECT_ID": self.project["id"],
            "SLAI_PROJECT_NAME": project_name,
        }

        return variables

    def _write_template_file(self, *, project_files, path, filename):
        with open(f"{path}/{filename}", "w") as f_out:
            f_out.write(project_files[filename])

    def _create_project_files(self, *, local_path):
        project_files = {}
        template_files = ["config.yml", "docker-compose.yml"]

        pwd = Path(__file__).parent

        # populate serverless project files
        for fname in template_files:
            log.action(f"Generating: {fname} ")
            template_contents = None

            with open(f"{pwd}/templates/{fname}", "r") as f_in:
                template_contents = f_in.read()
                t = Template(template_contents)
                rendered_template = t.render(**self.project_variables)

                project_files[fname] = rendered_template
                log.action("Done.")

        # write populated template files
        self._write_template_file(
            project_files=project_files,
            path=f"{local_path}/.slai",
            filename="config.yml",
        )

        self._write_template_file(
            project_files=project_files,
            path=f"{local_path}",
            filename="docker-compose.yml",
        )

    def _clone_template_repo(self, local_path):
        log.action("Cloning template repository")

        cloned = False
        for repo_url in TEMPLATE_REPO_URLS:

            try:
                Repo.clone_from(repo_url, local_path)
                rmtree(f"{local_path}/.git")
                cloned = True
                log.action("Done.")
            except git.exc.GitCommandError:
                pass

            if cloned:
                break

        if not cloned:
            log.warn("Unable to clone template repository.")
            raise RuntimeError("unable_to_clone_template_repo")

        return local_path



import slai
import torch


class ModelHandler(slai.base_handler):
    def model_inputs(self):
        return {"input_1": slai.inputs.Float, "input_2": slai.inputs.Float}

    def input(self, **inputs):
        x = torch.tensor([inputs["input_1"], inputs["input_2"]])
        return x

    def output(self, model_output):
        processed_output = model_output.detach().numpy().tolist()
        return processed_output



import docker
import sys
import threading
import time
import queue
import socket
import ast
import subprocess

from select import select
from slai_cli import log
from slai_cli.exceptions import RetryException

DOCKER_SOCKET_READ_SIZE = 1024
DOCKER_INPUT_UPDATE_RATE = 0.2


def _socket_input_worker(input_queue, stop_event):
    while not stop_event.is_set():
        # TODO: this apparently will not work on windows, need to find a solution for that
        _input, _, _ = select([sys.stdin], [], [], DOCKER_INPUT_UPDATE_RATE)

        if _input:
            _line = sys.stdin.readline()
            for c in _line:
                input_queue.put(c)


def _decode_string(s):
    frame_indices = []
    idx = 0
    in_frame = False
    current_start_index = 0

    while idx < len(s):
        if not in_frame:
            if s[idx] == 0x01:
                in_frame = True
                current_start_index = idx

        elif in_frame:
            if s[idx] != 0x00:
                frame_length = idx - current_start_index

                if frame_length == 8:
                    frame_indices.append([current_start_index, idx])
                    in_frame = False

        idx = idx + 1

    s = bytearray(s)
    initial_length = len(s)
    for frame in frame_indices:
        offset = len(s) - initial_length
        del s[frame[0] + offset : frame[1] + offset]

    return s.decode("utf-8", "ignore")


def _socket_output_worker(s, output_queue, stop_event):
    while not stop_event.is_set():
        try:
            msg = s._sock.recv(DOCKER_SOCKET_READ_SIZE)
            msg = _decode_string(msg)
            output_queue.put(msg)
        except socket.timeout:
            pass


class DockerClient:
    def __init__(self, *, project_name, project_id):
        self.project_name = project_name
        self.project_id = project_id
        try:
            self.client = docker.from_env()
            self.docker_api_client = docker.APIClient(
                base_url="unix://var/run/docker.sock"
            )  # TODO: support other platforms
        except docker.errors.DockerException:
            raise RuntimeError("unable_to_launch_container")

        self._connect_to_container()

    def _connect_to_container(self):
        self.project_container = None
        containers = self.client.containers.list(
            filters={"name": f"slai-{self.project_name}-{self.project_id}-workspace"}
        )

        if len(containers) == 1:
            self.project_container = containers[0]
            self.project_container_id = self.project_container.id
            self.host_port = self.project_container.ports["8888/tcp"][0]["HostPort"]
        else:
            log.warn("Unable to connect to docker container, attempting to start...")
            try:
                _ = subprocess.check_call(
                    "docker-compose up -d",
                    shell=True,
                )
                self._connect_to_container()
            except subprocess.CalledProcessError:
                raise RuntimeError("unable_to_launch_container")

    def _start_socket_io_worker(self, worker_function, args=()):
        _thread = threading.Thread(target=worker_function, args=args)
        _thread.daemon = True
        _thread.start()

    def create_model_environment(self, *, model_name):
        log.info(f"Checking if an environment exists for model: {model_name}...")

        _, output = self.project_container.exec_run(
            f"python /workspace/bin/create.py {model_name}",
        )

        if "exists" in output.decode("utf-8"):
            log.action(f"Using existing virtual environment for model: {model_name}")
        else:
            log.action(f"Created new virtual environment for model: {model_name}")

    def run_trainer(self, *, model_name, model_version_id):
        exit_code = self._run_interactive_command(
            cmd=f"/workspace/env/{model_name}/bin/python /workspace/models/{model_name}/{model_version_id}/trainer.py"  # noqa
        )
        return exit_code

    def test_model(self, *, model_name, model_version_id):
        exit_code = self._run_interactive_command(
            cmd=f"/workspace/env/{model_name}/bin/python /workspace/bin/setup_test_environment.py {model_name} {model_version_id}"  # noqa
        )
        if exit_code != 0:
            return exit_code

        exit_code = self._run_interactive_command(
            cmd=f"/workspace/env/{model_name}/bin/python /workspace/bin/test_model.py {model_name} {model_version_id}"  # noqa
        )

        return exit_code

    def install_requirement(self, *, model_name, requirement):
        exit_code = self._run_interactive_command(
            cmd=f"/workspace/env/{model_name}/bin/python -m pip install {requirement}"
        )
        return exit_code

    def check_model_requirements(self, *, model_name, requirements):
        log.action("Verifying artifact requirements...")

        req_string = " ".join(
            [f"{req}{version}" for req, version in requirements.items()]
        )
        exit_code = self._run_interactive_command(
            cmd=f"/workspace/env/{model_name}/bin/python -m pip install {req_string}"
        )
        return exit_code

    def check_trainer_imports(self, *, model_name, model_version_id):
        trainer_path = f"/workspace/models/{model_name}/{model_version_id}/trainer.py"

        exit_code = self._run_interactive_command(
            cmd=f"/workspace/env/{model_name}/bin/python /workspace/bin/check_imports.py {trainer_path}"  # noqa
        )
        return exit_code

    def generate_model_handler_requirements(self, *, model_name):
        _, output = self.project_container.exec_run(
            f"/workspace/env/{model_name}/bin/python /workspace/bin/generate_handler_requirements.py {model_name}",  # noqa
        )
        model_handler_requirements = ast.literal_eval(output.decode("utf-8"))
        return model_handler_requirements

    def check_handler_imports(self, *, model_name):
        handler_path = f"/workspace/handlers/{model_name}/handler.py"
        exit_code = self._run_interactive_command(
            cmd=f"/workspace/env/{model_name}/bin/python /workspace/bin/check_imports.py {handler_path}"  # noqa
        )
        return exit_code

    def launch_local_notebook_server(self, *, model_name):
        _, output = self.project_container.exec_run(
            "fuser -k 8888/tcp"
        )  # kill any running notebook servers

        killed_process = output.decode("utf-8").strip()
        if killed_process:
            log.action(f"Stopped existing notebook server: {killed_process}")

        try:
            _, output = self.project_container.exec_run(
                f"/workspace/env/{model_name}/bin/python -m jupyter notebook --allow-root --ip='0.0.0.0' --NotebookApp.token='' --NotebookApp.password='' --no-browser",  # noqa
                detach=True,
            )
        except docker.errors.APIError as e:
            if "python: no such file" in e.response.json()["message"]:
                log.warn(
                    "No virtual environment found for this model, checking environment..."
                )
                self.create_model_environment(model_name=model_name)
                raise RetryException
            else:
                return False

        return True

    def _run_interactive_command(self, *, cmd, silent=False):
        input_queue = queue.Queue()
        output_queue = queue.Queue()
        stop_event = threading.Event()

        exec_id = self.docker_api_client.exec_create(
            self.project_container_id,
            cmd,
            tty=True,
            stdin=True,
        )

        s = self.docker_api_client.exec_start(exec_id, socket=True)
        self._start_socket_io_worker(
            _socket_input_worker, args=(input_queue, stop_event)
        )
        self._start_socket_io_worker(
            _socket_output_worker, args=(s, output_queue, stop_event)
        )

        last_update = time.time()
        to_send = None

        process_info = self.docker_api_client.exec_inspect(exec_id)
        process_running = process_info["Running"]

        while process_running:
            if time.time() - last_update > DOCKER_INPUT_UPDATE_RATE:
                sys.stdout.flush()
                last_update = time.time()

            if not input_queue.empty():
                to_send = input_queue.get()
                s._sock.send(to_send.encode("utf-8"))
                to_send = None

            if not output_queue.empty():
                resp = output_queue.get()
                if not silent:
                    sys.stdout.write(resp)
                    sys.stdout.flush()

            process_info = self.docker_api_client.exec_inspect(exec_id)
            process_running = process_info["Running"]

        stop_event.set()
        s.close()

        if not silent:
            sys.stdout.write("\n")
            sys.stdout.flush()

        return process_info["ExitCode"]
